const std = @import("std");
const testing = @import("./testing.zig");

// TODO: Remove stdlib dependency.
pub const IntFittingRange = std.math.IntFittingRange;

///
/// Returns the highest integer value representable by `Integer`.
///
pub fn maxIntValue(comptime Integer: type) comptime_int {
    return switch (@typeInfo(Integer)) {
        .Int => |info| if (info.bits == 0) 0 else
            ((1 << (info.bits - @boolToInt(info.signedness == .signed))) - 1),

        else => @compileError("`" ++ @typeName(Integer) ++ "` must be an int"),
    };
}

test "maxIntValue" {
    try testing.expect(maxIntValue(u8) == 255);
    try testing.expect(maxIntValue(i8) == 127);

    try testing.expect(maxIntValue(u16) == 65535);
    try testing.expect(maxIntValue(i16) == 32767);
}

///
/// Returns the highest `Number` value between `this` and `that`.
///
pub fn max(comptime Number: type, this: Number, that: Number) Number {
    return switch (@typeInfo(Number)) {
        .Int, .Float, .ComptimeInt, .ComptimeFloat => if (this > that) this else that,

        else => @compileError("`" ++ @typeName(Number) ++
            "` must be an int, float, comptime_int, or comptime_float"),
    };
}

test "max" {
    try testing.expect(max(f32, 0.1, 1.0) == 1.0);
    try testing.expect(max(f64, 1.0, 1.01) == 1.01);
    try testing.expect(max(u32, 35615, 2873) == 35615);
}

///
/// Returns the lowest `Number` value between `this` and `that`.
///
pub fn min(comptime Number: type, this: Number, that: Number) Number {
    return switch (@typeInfo(Number)) {
        .Int, .Float, .ComptimeInt, .ComptimeFloat => if (this < that) this else that,

        else => @compileError("`" ++ @typeName(Number) ++
            "` must be an int, float, comptime_int, or comptime_float"),
    };
}

test "min" {
    try testing.expect(min(f32, 0.1, 1.0) == 0.1);
    try testing.expect(min(f64, 1.0, 1.01) == 1.0);
    try testing.expect(min(u32, 35615, 2873) == 2873);
}
